from typing import Tuple, List, Optional
from scipy.special import expit
from pathlib import Path

from numpy import mean, ndarray, sum, newaxis, power, arccos, clip, pi, argmax, array, tanh, zeros, savetxt, loadtxt
from numpy.random import randn, permutation, seed
from numpy.linalg import norm

from windML.fewshot.direct_feedback_alignment.utils import ModelSettings
from windML.utils.classifier import Classifier

class DFAClassifier(Classifier):
    """
    Feedforward Neural Network Classifier
    trained using Direct Feedback Alignment or Back Propagation
    """
    def __init__(self, input_dimension:int, output_dimension:int, load_path:Optional[str]=None, direct_feedback_alignment:bool=True) -> None:
        self.direct_feedback_alignment = direct_feedback_alignment
        if load_path is None:
            self.LABELS = list()
            self.weights1 = randn(ModelSettings.HIDDEN_DIMENSION.value, input_dimension)
            self.weights2 = randn(output_dimension, ModelSettings.HIDDEN_DIMENSION.value)
        else:
            self.LABELS = loadtxt(f"{load_path}/LABELS.txt",dtype=str)
            self.weights1 = loadtxt(f"{load_path}/weights1.txt")
            self.weights2 = loadtxt(f"{load_path}/weights2.txt")
        if self.direct_feedback_alignment:
            seed(0)
            self.random_weights = randn(ModelSettings.HIDDEN_DIMENSION.value, output_dimension)
    
    def labels(self) -> List[str]:
        return self.LABELS

    def predict(self, input:ndarray) -> ndarray:
        return self._predict_batch(array([input]).T).T[0]

    def _predict_batch(self, inputs:ndarray) -> ndarray:
        return self._forward_pass(array(inputs))
    
    def classify(self, input:ndarray) -> List[str]:
        id = argmax(self.predict(input))
        return self.LABELS[id]

    def save(self,name:str) -> str:
        path = f"{Path.cwd()}/{name}"
        Path(path).mkdir(parents=True, exist_ok=True)
        savetxt(f"{path}/LABELS.txt",self.LABELS,fmt="%s")
        savetxt(f"{path}/weights1.txt",self.weights1)
        savetxt(f"{path}/weights2.txt",self.weights2)

    def train_autoencoder(self, train_inputs:List[ndarray], train_labels:List[str]) -> None:
        self.LABELS = sorted(set(train_labels))
        x = array(train_inputs).T
        y = array(self._one_hot_encode_batch(train_labels,self.LABELS)).T
        self.fit(x,y)

    def fit(self, train_inputs:ndarray, train_outputs:ndarray) -> Tuple[List[float],List[float]]:        
        _,dataset_size = train_inputs.shape
        n_batches = dataset_size//ModelSettings.BATCH.value

        training_error,angles = list(),list()
        for epoch in range(ModelSettings.EPOCHS.value):
            train_inputs_shuffled,train_outputs_shuffled = self._shuffle_data(dataset_size, train_inputs,train_outputs)
            error,angle = self._partial_fit(n_batches,train_inputs_shuffled,train_outputs_shuffled)
            training_error.append(error/dataset_size)
            angles.append(angle)
            print(f'\nLoss at epoch {epoch+1}: Training error: {training_error[-1]}%')
        return training_error,angles

    def _backward_pass(self, error:ndarray, inputs:ndarray) -> None:
        self.weights2 += self._dWeights(error, self.final_hidden)
        error_activation = self._dActivation(error)     
        self.weights1 += self._dWeights(error_activation, inputs)

    def _partial_fit(self,n_batches:int, train_inputs:ndarray,train_outputs:ndarray) -> Tuple[float,float]:
        error,angle = 0.,0.
        for batch in range(n_batches):
            train_inputs_sampled,train_outputs_sampled = self._sample_data(batch,train_inputs,train_outputs)
            e,angle = self._fit_batch(train_inputs_sampled, train_outputs_sampled)
            error += e
            print(f"\rbatch: {batch} - {100*round(batch/n_batches,2)}%",end="")
        return error, angle

    def _fit_batch(self, train_inputs:ndarray, train_outputs:ndarray) -> Tuple[float,float]:
        predicted_output_logits = self._forward_pass(train_inputs)
        self._backward_pass(
            error=predicted_output_logits - train_outputs, 
            inputs=train_inputs
        )
        return (
            self._train_error(predicted_output_logits,train_outputs),
            self._average_angle(predicted_output_logits - train_outputs) if self.direct_feedback_alignment else 0
        )

    def _average_angle(self, error:ndarray) -> float:
        c = self._c(error)        
        dHidden =  self._dHidden(error)
        Lk = self._Lk(dHidden, c) 
        return self._angle(Lk, c)
        
    def _dActivation(self,error:ndarray) -> ndarray:
        e = self.random_weights @ error if self.direct_feedback_alignment else self.weights2.T @ error
        return e *  ( 1 - tanh(self.activation1)**2 )

    def _dHidden(self, error:ndarray) -> ndarray:
        return mean( self.random_weights @ error , axis=1)[:, newaxis]

    def _c(self, error:ndarray) -> ndarray:
        activationInverse = expit( self.activation2 ) 
        activationInverseMultInverse = 1 - activationInverse
        a = activationInverse * activationInverseMultInverse
        e = error * a
        return mean( self.weights2.T @ e, axis=1)[:, newaxis]

    def _first_half_forward_pass(self, inputs:ndarray) -> ndarray:
        self.activation1 = self._project(
            inputs=inputs,
            weights=self.weights1,
        )
        return tanh(self.activation1)

    def _last_half_forward_pass(self,inputs:ndarray) -> ndarray:
        self.activation2 = self._project(
            inputs=inputs,
            weights=self.weights2,
        )
        return expit(self.activation2)
    
    def _forward_pass(self, inputs:ndarray) -> ndarray:
        self.final_hidden = self._first_half_forward_pass(inputs)
        return self._last_half_forward_pass(self.final_hidden)
        
    @staticmethod
    def _dWeights(error:ndarray, inputs:ndarray) -> ndarray:
        dWeights = -error @ inputs.T 
        return ModelSettings.LEARNING_RATE.value * dWeights

    @staticmethod
    def _angle(Lk:ndarray, c:ndarray) -> float:
        return arccos(clip(Lk*DFAClassifier._inverse(c), -1., 1.)) * 180/pi

    @staticmethod
    def _Lk(dHidden:ndarray, c:ndarray) -> ndarray:
        return ((dHidden.T @ c) * DFAClassifier._inverse(dHidden))[0, 0]

    @staticmethod
    def _project(inputs:ndarray, weights:ndarray) -> ndarray: 
        return weights @ inputs 
    
    @staticmethod
    def _inverse(x:ndarray) -> ndarray:
        return power(norm(x), -1)

    @staticmethod
    def _shuffle_data(dataset_size:int, train_inputs:ndarray, train_outputs:ndarray) -> Tuple[ndarray,ndarray]:
        shuffled_ids = permutation(dataset_size)
        return train_inputs[:, shuffled_ids], train_outputs[:, shuffled_ids]

    @staticmethod
    def _sample_data(batch_number:int, train_inputs_shuffled:ndarray, train_outputs_shuffled:ndarray) -> Tuple[ndarray,ndarray]:
        return (
            train_inputs_shuffled[:, batch_number*ModelSettings.BATCH.value:(batch_number+1)*ModelSettings.BATCH.value],
            train_outputs_shuffled[:, batch_number*ModelSettings.BATCH.value:(batch_number+1)*ModelSettings.BATCH.value]
        )

    @staticmethod
    def _train_error(predicted:ndarray, expected:ndarray) -> int:
        return sum(argmax(predicted, axis=0) != argmax(expected, axis=0))

    @staticmethod
    def _one_hot_encode_batch(targets:List[str],labels:List[str]) -> List[ndarray]:
        vector_size = len(labels)
        return list(map(lambda target: DFAClassifier._one_hot_encode(labels.index(target),vector_size),targets))
    
    @staticmethod
    def _one_hot_encode(index:int, vector_size:int) -> ndarray:
        vector = zeros(vector_size)
        vector[index] = 1
        return vector


from typing import List, Optional
from pathlib import Path

from numpy import ndarray, array, argmax, savetxt, loadtxt
from numpy.linalg import pinv 

from windML.utils.classifier import Classifier
from windML.utils.label_encoding import one_hot_encode_batch

class MFRClassifier(Classifier):
    """
    Matrix Factorisation
    Retrieval-based classifier
    """
    def __init__(self,load_path:Optional[str]=None) -> None:
        if load_path is None:
            self.LABELS = list()
            self.label_vectors = None
        else:
            self.LABELS = loadtxt(f"{load_path}/LABELS.txt",dtype=str)
            self.label_vectors = loadtxt(f"{load_path}/label_vectors.txt") 

    def predict(self, input:ndarray) -> ndarray:
        return input @ self.label_vectors

    def classify(self, input:ndarray) -> str:
        label_id = argmax(self.predict(input))
        return self.LABELS[label_id]

    def fit(self, inputs:List[ndarray], labels:List[str]) -> None:
        self.LABELS = sorted(set(labels))
        self.label_vectors = self._matrix_factorisation(
            input_vectors=inputs,
            output_vectors_one_hot=one_hot_encode_batch(labels,self.LABELS),
        )

    def save(self,name:str) -> str:
        path = f"{Path.cwd()}/{name}"
        Path(path).mkdir(parents=True, exist_ok=True)
        savetxt(f"{path}/LABELS.txt",self.LABELS,fmt="%s")
        savetxt(f"{path}/label_vectors.txt",self.label_vectors)

    @staticmethod 
    def _matrix_factorisation(input_vectors:List[ndarray], output_vectors_one_hot:List[ndarray]) -> ndarray:    
        return pinv(input_vectors) @ array(output_vectors_one_hot)


from typing import Dict

from numpy import array, exp, ndarray, sqrt
from numpy.random import randint

class MultiHeadAttention:
    def __init__(self) -> None:
        pass

    def forward(self, Q:Dict[int,callable], K: Dict[int,callable], V: Dict[int,callable]) -> ndarray:
        attention_matrix = list()
        for index in Q.keys():
            logits = Q[index] @ K[index].T / sqrt(K[index].shape[1]) 
            outputs = self.softmax(logits)  
            attention_matrix.append(outputs @ V[index])

        attention_matrix = array(attention_matrix) 
        attention_heads = attention_matrix.reshape(self._permute_shape(attention_matrix)) 
        attention = attention_heads.reshape(attention_heads.shape[0], -1)  

        weights = randint(3, size=(attention.shape[-1], attention_matrix.shape[-1])) 
        return attention @ weights   

    def backward(self) -> None:
        pass

    @staticmethod
    def softmax(inputs: ndarray) -> ndarray:
        e = exp(inputs - max(inputs))
        return e / e.sum(axis=-1)

    @staticmethod
    def _permute_shape(inputs: ndarray) -> tuple:
        if len(inputs.shape) > 3:
            batch, heads, sequence_length, embeddings_size = inputs.shape
            return batch, sequence_length, heads, embeddings_size
        heads, sequence_length, embedding_size = inputs.shape
        return sequence_length, heads, embedding_size

    
    # def softmax_backward(self,error:ndarray) -> ndarray:
    #     return self.output * (error - sum( self.output * error, axis=-1, keepdims=True) )

    # def backward(self,inputs:ndarray, error:ndarray) -> Tuple[ndarray,ndarray,ndarray]:
    #     dConcatenatedOutput = self.weights.backward(inputs,error)
    #     dOutput = self._dOutput(dConcatenatedOutput)
    #     dScore = self._dScore(dOutput)

    #     dQuery = self._dQuery(dScore)
    #     dKey = self._dKey(dScore)
    #     dValue = self._dValue(dOutput)

    #     dQueries = self.weights_query.backward(inputs,dQuery).sum(axis=1,keepdims=True)
    #     dKeys = self.weights_keys.backward(inputs,dKey).sum(axis=1,keepdims=True)
    #     dValues = self.weights_values.backward(inputs,dValue).sum(axis=1,keepdims=True)        
    #     return dQueries,dKeys,dValues
    
    # def _dOutput(self, dConcatenatedOutput:ndarray) -> ndarray:
    #     return dConcatenatedOutput.reshape(-1,dConcatenatedOutput.shape[2],ModelSettings.N_HEADS.value,self.OUTPUT_DIMENSION).transpose(0,2,1,3)
    
    # def _dScore(self, dOutput:ndarray) -> ndarray:
    #     dScore = dOutput@self.values.transpose(0,1,3,2)
    #     dScore = self.softmax.backward(dScore)
    #     dScore *= sqrt(ModelSettings.HIDDEN_DIMENSION.value)
    #     return dScore
    
    # def _dValue(self, dOutput:ndarray) -> ndarray:
    #     return self.scores.transpose(0,1,3,2) @ dOutput

    # def _dQuery(self, dScore:ndarray) -> ndarray:
    #     return dScore @ self.keys
    
    # def _dKey(self, dScore:ndarray) -> ndarray:
    #     return dScore.transpose(0,1,-1,-2) @ self.queries

    # @staticmethod
    # def _scaled_scores(queries:ndarray, keys:ndarray) -> ndarray:
    #     scores = queries @ keys.transpose(0,1,-1,-2) 
    #     scores /= sqrt(ModelSettings.HIDDEN_DIMENSION.value) 
    #     return scores

    # @staticmethod
    # def _mask(shape:tuple) -> ndarray:
    #     return tile(triu(ones(shape[-1],shape[-1]),1) * ModelSettings.ATTENTION_MASK_CONSTANT.value,(shape[0],shape[1],1,1))



from typing import Optional, List
from pathlib import Path

from numpy import ndarray, exp, savetxt, loadtxt, argmax
from numpy.linalg import pinv
from numpy.random import uniform, seed

from windML.utils.label_encoding import one_hot_encode_batch

class ELMClassifier:
    """
    Extreme Learning Machine
    """
    def __init__(self, input_dimension:int, hidden_dimension:int=1000, load_path:Optional[str]=None) -> None:
        seed(0)
        self.random_projection = uniform(
            low=-.1, high=.1, 
            size =(input_dimension, hidden_dimension)
        )
        if load_path is None:
            self.LABELS = list()
            self.weights = None
        else:
            self.LABELS = loadtxt(f"{load_path}/LABELS.txt",dtype=str)
            self.weights = loadtxt(f"{load_path}/weights.txt")

    def fit(self, inputs: List[ndarray], labels: List[int]) -> None:
        self.LABELS = sorted(set(labels))
        label_vectors = one_hot_encode_batch(labels,self.LABELS)
        hidden = self._hidden_layer(inputs)
        self.weights = self._learn_dense_weights(hidden, label_vectors)

    def save(self, name:str) -> str:
        path = f"{Path.cwd()}/{name}"
        Path(path).mkdir(parents=True, exist_ok=True)
        savetxt(f"{path}/LABELS.txt",self.LABELS,fmt="%s")
        savetxt(f"{path}/weights.txt",self.weights)

    def predict(self, inputs: ndarray) -> ndarray:
        return self._output_layer(self._hidden_layer(inputs))
    
    def classify(self, inputs: ndarray) -> str:
        class_id = argmax(self.predict(inputs))
        return self.LABELS[class_id]

    def _hidden_layer(self, inputs: ndarray) -> ndarray: 
        return self.activation_function(inputs @ self.random_projection)
  
    def _output_layer(self, hidden: ndarray) -> ndarray: 
        return hidden @ self.weights
    
    def _learn_dense_weights(self, hidden:ndarray, label_vectors:List[ndarray]) -> ndarray:
        return pinv(hidden) @ label_vectors

    @staticmethod
    def activation_function(x: ndarray) -> ndarray: 
        return ELMClassifier.sigmoid(x)
    
    @staticmethod
    def sigmoid(x: ndarray) -> ndarray:
        return 1. / (1. + exp(-x))


from enum import Enum 

class ModelSettings(Enum):
    SEQUENCE_LENGTH = 15
    EMBEDDING_DIMENSION = 128
    VOCABULARY_SIZE = 128
    N_HEADS = 8
    N_DECODERS = 8
    SCALING_FACTOR = .01
    DROPOUT_RATE = .2
    AVOID_DIVISION_BY_ZERO = 1e-5

assert ModelSettings.EMBEDDING_DIMENSION.value % ModelSettings.N_HEADS.value == 0


from numpy import ndarray
from numpy.random import rand

from windML.fewshot.transformer.utils import ModelSettings

class DropoutLayer:
    def __init__(self) -> None:
        pass

    def dropout_layer(self, inputs: ndarray) -> ndarray:
        probability = 1. - ModelSettings.DROPOUT_RATE.value
        keep = rand(*inputs.shape) < probability
        outputs = inputs * keep
        outputs /= probability
        return outputs

    def backward(self) -> None:
        pass



from windML.oneshot.models.extreme_learning_machine import ELMClassifier
from windML.oneshot.modelfree.matrix_factorisation import MFRClassifier
from windML.fewshot.direct_feedback_alignment.dfa_classifier import DFAClassifier
from windML.fewshot.recurrent_neural_network.character_rnn import charRNN
from windML.fewshot.recurrent_neural_network.rnn import RNN

__version__ = "0.2.3"



from lib2to3.pgen2 import token
from numpy.random import randint

from windML.fewshot.transformer.model import Transformer
from windML.fewshot.transformer.utils import ModelSettings

model = Transformer()

token_ids = randint(3, size=(2, ModelSettings.SEQUENCE_LENGTH.value, ModelSettings.VOCABULARY_SIZE.value))
print(token_ids)
predictions = model.forward(token_ids[0,:,:], token_ids[1,:,:])
print(predictions)


from numpy import ndarray, sqrt, ones, zeros, mean

from windML.fewshot.transformer.utils import ModelSettings

class NormalisationLayer:
    def __init__(self) -> None:
        pass

    def forward(self, inputs: ndarray) -> ndarray:
        normalisation_dimensions = inputs.shape[1:] if len(inputs.shape) > 3 else inputs.shape
        averages = mean(inputs, keepdims=True)  
        variance = mean(inputs**2, keepdims=True) - averages ** 2
        normalised = (inputs - averages) / sqrt(variance + ModelSettings.AVOID_DIVISION_BY_ZERO.value) 
        gain = ones(shape=normalisation_dimensions) 
        bias = zeros(shape=normalisation_dimensions) 
        return gain * normalised + bias

    # def backward(self,error:ndarray) -> ndarray:
    #     batch_size = self.normalised.shape[-1]
    #     normalised_inputs = self.inputs-mean(self.inputs,axis=-1,keepdims=True)
    #     normalised_error = normalised_inputs*error
    #     total_normalised_error = normalised_error.sum(axis=-1,keepdims=True)
    #     total_error = error.sum(axis=-1,keepdims=True)
    #     batch_error = batch_size*error-total_error
    #     a = sqrt(self.variance)*batch_error-self.norm*total_normalised_error
    #     b = batch_size*self.variance
    #     return a / b

    # @staticmethod
    # def _normalise(inputs:ndarray, variance:ndarray) -> ndarray:
    #     return ( inputs - mean(inputs,axis=-1,keepdims=True) ) / sqrt(variance)



from numpy import ndarray, broadcast_to
from numpy.random import randint 

from windML.fewshot.transformer.utils import ModelSettings

class DenseLayer:
    def __init__(self) -> None:
        pass 

    def forward(self, inputs: ndarray) -> ndarray:
        weights = randint(2, size=(inputs.shape[-1], ModelSettings.EMBEDDING_DIMENSION.value)) * ModelSettings.SCALING_FACTOR.value
        bias = randint(2, size=(ModelSettings.EMBEDDING_DIMENSION.value))
        outputs = inputs @ weights 
        outputs += broadcast_to(bias, outputs.shape)
        return outputs
    
    # def backward(self,inputs:ndarray,error:ndarray):
    #     dWeights = self._dWeights(inputs,error)            
    #     dInputs = self._dInputs(error)
    #     self.weights -= dWeights*ModelSettings.LEARNING_RATE.value
    #     dBias = None if self.bias is None else self._dBias(error)
    #     self.adam(dWeights, dBias)
    #     return dInputs
    
    # def adam(self, dWeights:ndarray, dBias:Optional[ndarray]) -> None:
    #     if self.vWeights is None:
    #         self.vWeights = dWeights
    #         self.sWeights = dWeights**2
            
    #     if dBias is not None:
    #         if self.vBias is None:
    #             self.vBias = dBias
    #             self.sBias = dBias**2
            
    #         self.vBias = self._vValue(self.vBias, dBias)
    #         self.sBias = self._sValue(self.sBias, dBias)
    #         self.bias -= self._dValue(self.vBias, self.sBias)
            
    #     self.vWeights = self._vValue(self.vWeights,dWeights)
    #     self.sWeights = self._sValue(self.sWeights, dWeights)
    #     self.weights -= self._dValue(self.vWeights, self.sWeights)

    # def _dInputs(self, error) -> ndarray:
    #     return error @ self.weights.transpose(0,-1,1)

    # @staticmethod
    # def _initialise_weights(n_heads:int,input_dimension:ndarray,output_dimension:ndarray) -> ndarray:
    #     weights = randn(n_heads,input_dimension,output_dimension) 
    #     weights *= sqrt(2/(n_heads*input_dimension))     
    #     return weights

    # @staticmethod
    # def _dWeights(inputs:ndarray, error:ndarray) -> ndarray:
    #     dWeights = inputs.transpose(0,1,-1,-2) @ error
    #     return dWeights.sum(axis=0)

    # @staticmethod
    # def _dBias(error:ndarray) -> ndarray:
    #     return error.sum(axis=0,keepdims=True).sum(axis=2,keepdims=True)

    # @staticmethod
    # def _vValue(vValue:ndarray, dValue:ndarray) -> ndarray:
    #     return vValue*ModelSettings.ADAM_ALPHA.value + dValue*(1-ModelSettings.ADAM_ALPHA.value)

    # @staticmethod
    # def _sValue(sValue:ndarray, dValue:ndarray) -> ndarray:
    #     return sValue*ModelSettings.ADAM_BETA.value + (dValue**2)*(1-ModelSettings.ADAM_BETA.value)

    # @staticmethod
    # def _dValue(vValue:ndarray, sValue:ndarray) -> ndarray:
    #     return vValue*ModelSettings.LEARNING_RATE.value / ( sqrt(sValue) + ModelSettings.ADAM_CONSTANT.value )



from numpy import ndarray, arange, newaxis, cos, sin, power

from windML.fewshot.transformer.utils import ModelSettings


class PositionEncodingLayer:
    def __init__(self) -> None:
        pass

    def forward(self, inputs: ndarray) -> ndarray:
        sequence_length = inputs.shape[0]
        position_encodings = self._position_encodings(
            positions=arange(sequence_length)[:, newaxis],
            indexes=arange(ModelSettings.EMBEDDING_DIMENSION.value)[newaxis, :]
        )
        if len(inputs.shape) <= 1: 
            position_encodings = position_encodings[newaxis,...]
        return inputs + position_encodings
        
    def backward(self) -> None:
        pass

    @staticmethod
    def _position_encodings(positions:ndarray, indexes:ndarray):
        angles = 1 / power(10000., (2*(indexes//2)) / ModelSettings.EMBEDDING_DIMENSION.value)
        angles = positions * angles
        angles[:, 0::2] = sin(angles[:, 0::2])
        angles[:, 1::2] = cos(angles[:, 1::2])
        return angles


from typing import List, Dict

from numpy import ndarray, add

from windML.fewshot.transformer.utils import ModelSettings
from windML.fewshot.transformer.layers.dense import DenseLayer
from windML.fewshot.transformer.layers.position_encoder import PositionEncodingLayer
from windML.fewshot.transformer.layers.multihead_attention import MultiHeadAttention
from windML.fewshot.transformer.layers.residual import ResidualLayer
from windML.fewshot.transformer.layers.dropout import DropoutLayer
from windML.fewshot.transformer.layers.normalisation import NormalisationLayer


class DecoderBlock:
    def __init__(self) -> None:
        self.positional_encoding_layer = PositionEncodingLayer() 
        self.multihead_attention1 = MultiHeadAttention()
        self.multihead_attention2 = MultiHeadAttention()
        self.residual_layer1 = ResidualLayer()
        self.residual_layer2 = ResidualLayer()
        self.residual_layer3 = ResidualLayer()
        self.dropout_layer1 = DropoutLayer()
        self.dropout_layer2 = DropoutLayer()
        self.dropout_layer3 = DropoutLayer()
        self.normalisation_layer1 = NormalisationLayer()
        self.normalisation_layer2 = NormalisationLayer()
        self.normalisation_layer3 = NormalisationLayer()

    def forward(self, context: ndarray, inputs: ndarray) -> ndarray:
        context = self.positional_encoding_layer.forward(context)        

        attention1 = self.multihead_attention1.forward(
            Q=self.deep_layers(context), 
            K=self.deep_layers(context), 
            V=self.deep_layers(context)
        )
        summed = add(context, attention1)
        residual1 = self.residual_layer1.forward(summed, output_dimension=summed.shape[-1])
        dropout1 = self.dropout_layer1.forward(residual1)
        normalised1 = self.normalisation_layer1.forward(dropout1)

        attention2 = self.multihead_attention2.forward(
            Q=self.deep_layers(context),
            K=self.deep_layers(inputs),
            V=self.deep_layers(inputs)
        )
        summed2 = add(normalised1, attention2)
        residual2 = self.residual_layer2.forward(summed2, output_dimension=summed2.shape[-1])
        dropout2 = self.dropout_layer2.forward(residual2)
        normalised2 = self.normalisation_layer2.forward(dropout2)

        residual3 = self.residual_layer3.forward(normalised2, output_dimension=normalised2.shape[-1])
        dropout3 = self.dropout_layer3.forward(residual3)
        normalised3 = self.normalisation_layer3.forward(dropout3)
        return normalised3

    @staticmethod
    def deep_layers(inputs:ndarray) -> List[callable]:
        return dict(map(
            lambda index:(index, DenseLayer().forward(inputs)),
            range(ModelSettings.N_HEADS.value)
        ))


    # def backward(self,inputs:ndarray,error:ndarray) -> Tuple[ndarray,ndarray]:
    #     dLayer3 = self.normalisation_layer3.backward(error)
    #     dLogits = self.dense_layer2.backward(inputs,dLayer3)
    #     dLogits[self.activations] = 0
    #     dNormLayer2 = self.dense_layer1.backward(inputs,dLogits) + dLayer3
    #     dLayer2 = self.normalisation_layer2.backward(dNormLayer2)        
    #     dNormLayer1, dNormLayer1Key, dNormLayer1Value = self.attention_layer2.backward(inputs,dLayer2)  
    #     dNormLayer1 += dLayer2
    #     dLayer1 = self.normalisation_layer1.backward(dNormLayer1)        
    #     dInput, dInputKey, dInputValue = self.attention_layer1.backward(inputs,dLayer1)        
    #     dInput = dInput + dInputKey + dInputValue + dLayer1
    #     dContext = dNormLayer1Key + dNormLayer1Value
    #     return dInput, dContext



from typing import List 

from numpy import zeros, ndarray

def one_hot_encode_batch(targets:List[str],labels:List[str]) -> List[ndarray]:
    vector_size = len(labels)
    return list(map(lambda target: one_hot_encode(labels.index(target),vector_size),targets))
    
def one_hot_encode(index:int, vector_size:int) -> ndarray:
    vector = zeros(vector_size,dtype=int)
    vector[index] = 1
    return vector



from setuptools import setup, find_packages

VERSION = '0.2.3'
DESCRIPTION = 'WindML'
LONG_DESCRIPTION = 'Refreshingly Lightweight Machine Learning (using only numpy). models include conditional-character-RNN, Transformer and a Direct Feedback Alignment NN and other one-shot models like extreme learning machines or matrix factorisation retrieval-based classifiers (for a model-free alternative)'

setup(
    name="windML",
    version=VERSION,
    author="Mohammed Terry-Jack",
    author_email="<mohammedterryjack@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['numpy'],
    keywords=['python', 'extreme learning machine', 'direct feedback alignment', 'matrix factorisation', 'ml', 'machine learning', 'rnn', 'transformer', 'numpy', 'lite'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
    ]
)


from typing import List
from numpy import ndarray

class Classifier:
    def labels(self) -> List[str]:
        raise NotImplementedError

    def predict(self, inputs:ndarray) -> ndarray:
        raise NotImplementedError

    def classify(self, inputs:ndarray) -> List[str]:
        raise NotImplementedError
    
    def save(self, name:str) -> str:
        raise NotImplementedError
    
    def fit(self, inputs:List[ndarray], targets:List[int]) -> None:
        raise NotImplementedError



from typing import List, Tuple, Optional
from pathlib import Path

from numpy.random import randn, choice
from numpy import argmax, zeros, ndarray, copy, tanh, exp, sum, log, zeros_like, clip, sqrt, savetxt, loadtxt

from windML.fewshot.recurrent_neural_network.utils import ModelSettings, VOCABULARY
from windML.utils.label_encoding import one_hot_encode


class charRNN:
    """
    character-level Recurrent Neural Network (RNN)
    """
    def __init__(self, load_path:Optional[str]=None, hidden_dimension:int=ModelSettings.HIDDEN_DIMENSION.value) -> None:
        if load_path is None:
            self.weights1 = randn(hidden_dimension, ModelSettings.VOCABULARY_SIZE.value)*ModelSettings.WEIGHT_SCALE_FACTOR.value     
            self.weights2 = randn(hidden_dimension, hidden_dimension)*ModelSettings.WEIGHT_SCALE_FACTOR.value 
            self.weights3 = randn(ModelSettings.VOCABULARY_SIZE.value, hidden_dimension)*ModelSettings.WEIGHT_SCALE_FACTOR.value
            self.bias1 = zeros((hidden_dimension, 1)) 
            self.bias2 = zeros((ModelSettings.VOCABULARY_SIZE.value, 1)) 
        else:
            self.weights1 = loadtxt(f"{load_path}/weights1.txt")
            self.weights2 = loadtxt(f"{load_path}/weights2.txt")
            self.weights3 = loadtxt(f"{load_path}/weights3.txt")
            self.bias1 = loadtxt(f"{load_path}/bias1.txt")
            self.bias2 = loadtxt(f"{load_path}/bias2.txt")
            self.bias1 = self.bias1.reshape((self.bias1.shape[0],1))
            self.bias2 = self.bias2.reshape((self.bias2.shape[0],1))
        self.hidden_dimension = hidden_dimension    
    
    def save(self, name:str) -> None:
        path = f"{Path.cwd()}/{name}"
        Path(path).mkdir(parents=True, exist_ok=True)
        savetxt(f"{path}/weights1.txt",self.weights1)
        savetxt(f"{path}/weights2.txt",self.weights2)
        savetxt(f"{path}/weights3.txt",self.weights3)
        savetxt(f"{path}/bias1.txt",self.bias1)
        savetxt(f"{path}/bias2.txt",self.bias2)

    def generate(self, prompt:str, condition_vector:Optional[ndarray]=None, generation_length:int=ModelSettings.SEQUENCE_LENGTH.value, greedy:bool=True) -> str:
        if condition_vector is None:
            hidden_context = self._new_hidden_context()
        else: 
            assert condition_vector.shape == (self.hidden_dimension,1)
            hidden_context = condition_vector 

        token_ids = [self.tokenise_character(ModelSettings.BOS_TOKEN.value)]

        for character in prompt:
            encoded_input = self._input(token_ids[-1])
            hidden_context = self._hidden_state(encoded_input,hidden_context)
            token_ids.append(self.tokenise_character(character))

        for _ in range(generation_length):    
            encoded_input = self._input(token_ids[-1])     
            hidden_context = self._hidden_state(encoded_input,hidden_context)
            logits = self._logits(hidden_context)
            index = self._choose_index_greedily(logits) if greedy else self._choose_index_stochastically(logits)
            if index == ModelSettings.VOCABULARY_SIZE.value-1:
                break
            token_ids.append(index)

        return self.decode_token_ids(token_ids)

    def fit(self, sentences:List[str], encoded_contexts:Optional[List[ndarray]]=None, epochs:int=ModelSettings.EPOCHS.value) -> None:
        token_ids = list(map(self.tokenise_string,sentences))
        encoded_inputs = list(map(self.encode_token_ids,token_ids))
        if encoded_contexts is None:
            encoded_contexts = list(map(lambda _:None,encoded_inputs))
        else:
            assert encoded_contexts[0].shape == (self.hidden_dimension,1)
        mWeights1 = zeros_like(self.weights1)
        mWeights2 = zeros_like(self.weights2)
        mWeights3 = zeros_like(self.weights3)
        mBias1 = zeros_like(self.bias1)
        mBias2 = zeros_like(self.bias2) 

        for epoch in range(epochs):
            for context,inputs,targets in zip(encoded_contexts,encoded_inputs,token_ids):
                inputs = inputs[:-1]
                targets = targets[1:]                
                loss = self._forward_pass(inputs, targets, context)
                self._update_parameters(
                    parameters=(self.weights1, self.weights2, self.weights3, self.bias1, self.bias2),
                    dParameters=self._backward_pass(inputs, targets),
                    mParameters=(mWeights1, mWeights2, mWeights3, mBias1, mBias2)
                )            
            print(f'epoch {epoch} / {epochs} [loss = {loss}]') 

    def _forward_pass(self, inputs:List[ndarray], targets:List[int], context_vector:Optional[ndarray]=None) -> float:        
        self.inputs = dict()
        self.probabilities = dict()        
        self.hiddens = dict()
        self.hiddens[-1] = self._new_hidden_context() if context_vector is None else context_vector

        loss = 0
        for position,encoded_input in enumerate(inputs): 
            raw_input = self._input(encoded_input)
            hidden_state = self._hidden_state(raw_input, self.hiddens[position-1])
            logits = self._logits(hidden_state)
            probability = self.softmax(logits) 
            loss += self.cross_entropy(probability,targets[position])
            self.inputs[position] = raw_input
            self.hiddens[position] = hidden_state
            self.probabilities[position] = probability
        return loss

    def _backward_pass(self, inputs:List[ndarray], targets:List[int]) -> Tuple[ndarray,ndarray,ndarray,ndarray,ndarray]:
        dWeights1 = zeros_like(self.weights1)
        dWeights2 = zeros_like(self.weights2)
        dWeights3 = zeros_like(self.weights3) 
        dBias1 =  zeros_like(self.bias1)
        dBias2 = zeros_like(self.bias2)
        dHiddenContext = self._new_hidden_context()

        for position in reversed(range(len(inputs))):
            dTarget = self._dTarget(targets[position], self.probabilities[position])
            dWeights3 += self._dWeights3(dTarget,self.hiddens[position])
            dBias2 += dTarget 
            dHidden = self._dHidden(dTarget,dHiddenContext)
            dInput = self._dInput(self.hiddens[position], dHidden)
            dBias1 += dInput
            dWeights1 += self._dWeights1(self.inputs[position], dInput)
            dWeights2 += self._dWeights2(self.hiddens[position-1], dInput)
            dHiddenContext = self._dHiddenContext(dInput)

        for value in [dWeights1, dWeights2, dWeights3, dBias1, dBias2]: 
            clip(value, -ModelSettings.OVERFLOW_LIMIT.value, ModelSettings.OVERFLOW_LIMIT.value, out=value)
        return dWeights1,dWeights2,dWeights3,dBias1,dBias2
    
    def _hidden_state(self, input:ndarray, prior_hidden_state:ndarray) -> ndarray:
        return tanh(
            (self.weights1 @ input) + 
            (self.weights2 @ prior_hidden_state) + 
            self.bias1
        )

    def _logits(self, hidden_state:ndarray) -> ndarray:
        return self.weights3 @ hidden_state + self.bias2

    def _dHidden(self, dTarget:ndarray, dHiddenContext:ndarray) -> ndarray:
        return self.weights3.T @ dTarget + dHiddenContext

    def _dHiddenContext(self, dInput:ndarray) -> ndarray:
        return self.weights2.T @ dInput

    def _new_hidden_context(self) -> ndarray:
        return zeros((self.hidden_dimension,1)) 

    @staticmethod
    def cross_entropy(probability:ndarray, target:ndarray) -> float:
        return -log(probability[target,0]) 

    @staticmethod
    def softmax(x:float) -> float:
        e = exp(x)
        return e / sum(e) 

    @staticmethod
    def format_text(text:str) -> str:
        return ModelSettings.BOS_TOKEN.value + text.lower().replace(
            ModelSettings.BOS_TOKEN.value,ModelSettings.PAD_TOKEN.value
        ).replace(
            ModelSettings.EOS_TOKEN.value,ModelSettings.PAD_TOKEN.value
        ) + ModelSettings.EOS_TOKEN.value + ModelSettings.PAD_TOKEN.value*ModelSettings.SEQUENCE_LENGTH.value

    @staticmethod
    def tokenise_string(text:str) -> List[int]:
        return list(map(charRNN.tokenise_character, charRNN.format_text(text)))[:ModelSettings.SEQUENCE_LENGTH.value+1]

    @staticmethod
    def tokenise_character(character:str) -> int:
        return VOCABULARY.index(character) if character in VOCABULARY else 0

    @staticmethod
    def encode_token_ids(token_ids:List[int]) -> ndarray:
        return list(map(lambda id: one_hot_encode(id,ModelSettings.VOCABULARY_SIZE.value),token_ids))

    @staticmethod
    def decode_token_ids(token_ids:List[int]) -> str:
        return ''.join(map(lambda token_id:VOCABULARY[token_id], token_ids))

    @staticmethod
    def _dTarget(target:ndarray, probability:ndarray) -> ndarray:
        dTarget = copy(probability) 
        dTarget[target] -= 1 
        return dTarget

    @staticmethod
    def _dWeights3(dTarget:ndarray, hidden_state:ndarray) -> ndarray:
        return dTarget @ hidden_state.T

    @staticmethod
    def _dInput(hidden_state:ndarray, dHidden:ndarray) -> ndarray:
        return (1 - hidden_state ** 2 ) * dHidden

    @staticmethod
    def _dWeights1(input:ndarray, dInput:ndarray) -> ndarray:
        return dInput @ input.T

    @staticmethod
    def _dWeights2(prior_hidden_state:ndarray, dInput:ndarray) -> ndarray:
        return dInput @ prior_hidden_state.T

    @staticmethod
    def _input(encoded_input:ndarray) -> ndarray:
        input = zeros((ModelSettings.VOCABULARY_SIZE.value,1))
        input[encoded_input] = 1
        return input

    @staticmethod
    def _update_parameters(
        parameters:Tuple[ndarray,ndarray,ndarray,ndarray,ndarray],
        dParameters:Tuple[ndarray,ndarray,ndarray,ndarray,ndarray],
        mParameters:Tuple[ndarray,ndarray,ndarray,ndarray,ndarray]
    ) -> None:
        for value, dValue, mValue in zip(parameters, dParameters, mParameters):
            mValue += dValue * dValue 
            value -= ModelSettings.LEARNING_RATE.value * dValue / sqrt(mValue + ModelSettings.ADAGRAD.value)     

    @staticmethod
    def _choose_index_stochastically(logits:ndarray) -> int:
        return choice(
            a=range(ModelSettings.VOCABULARY_SIZE.value), 
            p=charRNN.softmax(logits).ravel()
        ) 

    @staticmethod
    def _choose_index_greedily(logits:ndarray) -> int:
        return argmax(logits)


from numpy import ndarray, ones, broadcast_to, maximum 
from numpy.random import randint

class ResidualLayer:
    def __init__(self) -> None:
        pass

    def forward(self, inputs: ndarray, output_dimension:int) -> ndarray:
        weights_residual = randint(5, size=(inputs.shape[-1], output_dimension))
        bias_residual = ones(shape=(output_dimension))
        outputs = inputs @ weights_residual
        outputs += inputs 
        outputs += broadcast_to(bias_residual, outputs.shape)
        return self.relu(outputs) 

    def backward(self) -> None:
        pass

    @staticmethod
    def relu(inputs: ndarray) -> ndarray:
        return maximum(inputs, 0, inputs)


from numpy import ndarray, exp, max, broadcast_to
from numpy.random import randint

from windML.fewshot.transformer.decoder_block import DecoderBlock
from windML.fewshot.transformer.utils import ModelSettings

class Transformer:
    def __init__(self) -> None: 
        self.decoder_stack = list(map(lambda _: DecoderBlock(),range(ModelSettings.N_DECODERS.value)))
        self.weights = randint(2, size=(ModelSettings.EMBEDDING_DIMENSION.value * ModelSettings.SEQUENCE_LENGTH.value, ModelSettings.VOCABULARY_SIZE.value)) * ModelSettings.SCALING_FACTOR.value
        self.bias = randint(2, size=(ModelSettings.VOCABULARY_SIZE.value))

    def forward(self, context:ndarray, inputs:ndarray) -> ndarray:
        for decoder in self.decoder_stack:
            inputs = decoder.forward(inputs, context)
        outputs = self.dense_layer(inputs.flatten())
        return self.softmax(outputs)

    def dense_layer(self, inputs: ndarray) -> ndarray:        
        outputs = inputs @ self.weights 
        outputs += broadcast_to(self.bias, outputs.shape)
        return Transformer.sigmoid(outputs) 
  
    @staticmethod
    def sigmoid(inputs:ndarray) -> ndarray:
        return 1/(1 + exp(-inputs))  

    @staticmethod
    def softmax(inputs: ndarray) -> ndarray:
        e = exp(inputs - max(inputs))
        return e / e.sum(axis=-1)


from enum import Enum 
from string import punctuation

VOCABULARY = " 0123456789abcdefghijklmnopqrstuvwxyz" + punctuation

class ModelSettings(Enum):
    VOCABULARY_SIZE = len(VOCABULARY)    
    HIDDEN_DIMENSION = 100 
    EPOCHS = 1000
    SEQUENCE_LENGTH = 30
    LEARNING_RATE = 1e-3
    WEIGHT_SCALE_FACTOR = 1e-2
    OVERFLOW_LIMIT = 5
    ADAGRAD = 1e-8
    BOS_TOKEN = "$"
    EOS_TOKEN = "~"
    PAD_TOKEN = " "